@@ -6,6 +6,7 @@ import json
 import threading
 import fnmatch
 from datetime import datetime
+from typing import Union
 
 from lark import Token
 from lark import Lark, Transformer
@@ -55,6 +56,7 @@ grammar = r"""
                | sum ">=" sum -> ge
                | sum "in" sum -> contains
                | sum "contains" sum -> contained_in
+               | sum "has" sum -> has_property
                | sum "like" sum -> pattern_match
                | sum "not" "like" sum -> pattern_not_match
                | sum "is" "$past" -> date_is_past
@@ -89,12 +91,12 @@ grammar = r"""
          | RESOLVER_EXPR -> resolver_expression
          | "(" expr ")" -> paren
 
-    ?array_quantified: "at" "least" INTEGER "of" expr "has" RESOLVER_EXPR      -> at_least_of
-                     | "at" "most"  INTEGER "of" expr "has" RESOLVER_EXPR      -> at_most_of
-                     | "exactly"    INTEGER "of" expr "has" RESOLVER_EXPR      -> exactly_of
-                     | "any"  "of" expr "has" RESOLVER_EXPR                    -> any_of_sugar
-                     | "all"  "of" expr "has" RESOLVER_EXPR                    -> all_of_sugar
-                     | "none" "of" expr "has" RESOLVER_EXPR                    -> none_of_sugar
+    ?array_quantified: "at" "least" INTEGER "of" expr ("match" | "matches") ARRAY_EXPR      -> at_least_of
+                     | "at" "most"  INTEGER "of" expr ("match" | "matches") ARRAY_EXPR      -> at_most_of
+                     | "exactly"    INTEGER "of" expr ("match" | "matches") ARRAY_EXPR      -> exactly_of
+                     | "any"  "of" expr ("match" | "matches") ARRAY_EXPR                    -> any_of_sugar
+                     | "all"  "of" expr ("match" | "matches") ARRAY_EXPR                    -> all_of_sugar
+                     | "none" "of" expr ("match" | "matches") ARRAY_EXPR                    -> none_of_sugar
 
     // Define reserved keywords
     // But use string literals in rules above for "or", "and", "True", "False"
@@ -102,10 +104,14 @@ grammar = r"""
 
     FUNC_NAME: /(count_of|any_of|all_of|none_of)/
 
-    VARIABLE: /(?!or\b|and\b|true\b|false\b|is\b|contains\b|like\b|in\b|count_of\b|any_of\b|all_of\b|none_of\b)[a-zA-Z_][a-zA-Z0-9_]*(?:'s\s+[a-zA-Z_][a-zA-Z0-9_]*|\.[a-zA-Z_][a-zA-Z0-9_]*|\[\d+\])*/
+    VARIABLE: /(?!or\b|and\b|true\b|false\b|is\b|contains\b|like\b|in\b|has\b|match\b|matches\b|count_of\b|any_of\b|all_of\b|none_of\b)[a-zA-Z_][a-zA-Z0-9_]*(?:'s\s+[a-zA-Z_][a-zA-Z0-9_]*|\.[a-zA-Z_][a-zA-Z0-9_]*|\[\d+\])*/
 
     func_call: FUNC_NAME "(" expr ("," RESOLVER_EXPR)? ")"
-    
+
+    // Array expression syntax: |expression| - must be matched as a single token
+    // Used for array quantified expressions to avoid conflicts with jq syntax
+    ARRAY_EXPR: /\|[^|]*\|/
+
     // JQ expression syntax: `expression` - must be matched as a single token
     // Define this before the STRING token to give it higher precedence
     RESOLVER_EXPR: /`[^`]*`/
@@ -275,6 +281,32 @@ class ExpressionTransformer(Transformer, DateMethods, ArrayMethods):
             container=left, item=right, container_position="contains"
         )
 
+    @binary_op
+    def has_property(self, left, right) -> bool:
+        """
+        Check if the left operand (object) has the right operand as a property/key.
+
+        This is similar to 'in' but with more natural syntax: object has property
+        instead of property in object.
+
+        Examples:
+        - user has 'name'      -> checks if user object has a 'name' key
+        - config has debug     -> checks if config object has a 'debug' key
+        - user has field_name  -> checks using variable field_name as key
+        """
+        if isinstance(left, dict):
+            # For dictionaries, check if the key exists
+            return right in left
+        elif isinstance(left, (list, tuple)):
+            # For lists/tuples, check if the item exists in the collection
+            return right in left
+        elif hasattr(left, right) and isinstance(right, str):
+            # For objects with attributes, check if attribute exists
+            return hasattr(left, right)
+        else:
+            # For other types, return False (they don't have properties)
+            return False
+
     @binary_op
     def pattern_match(self, left, right) -> bool:
         """
@@ -334,6 +366,22 @@ class ExpressionTransformer(Transformer, DateMethods, ArrayMethods):
 
         return value
 
+    def array_expression(
+        self, items: list[Token]
+    ) -> int | float | bool | str | list | dict | datetime:
+        """Process a pipe-delimited expression for array quantified operations"""
+        # Extract the expression from the token: |expression| -> expression
+        raw_expr = items[0].value[1:-1]  # Remove | prefix and | suffix
+
+        # Use the resolver system to evaluate with raw=True
+        value = resolve_path(raw_expr, self.processed_json, raw=True)
+
+        # Handle datetime reconstruction
+        if isinstance(value, dict) and "__datetime__" in value:
+            return datetime.fromisoformat(value["__datetime__"])
+
+        return value
+
     # Array quantified sugar methods
     def at_least_of(self, items):
         """Transform 'at least N of X has P' to count_of(X, P) >= N"""
@@ -391,23 +439,48 @@ class CompiledExpression:
         self.expression = expression
         self.parse_tree = parse_tree
 
-    def evaluate(self, variables: dict | str | None = None) -> int | float | bool | str:
+    def evaluate(
+        self, context: Union[dict, str, "ProcessedContext", None] = None
+    ) -> Union[int, float, bool, str]:
         """
-        Evaluate this compiled expression with the provided variables.
+        Evaluate this compiled expression with the provided context.
 
         Args:
-            variables: Dictionary or JSON string containing variable values
+            context: Dictionary, JSON string, or ProcessedContext containing variable values
 
         Returns:
             The result of evaluating the expression
         """
-        processed_json = _process_variables(variables)
+        processed_json = _extract_processed_json(context)
 
         with execution_error_handling(self.expression):
             transformer = ExpressionTransformer(processed_json=processed_json)
             return transformer.transform(self.parse_tree)
 
 
+class ProcessedContext:
+    """
+    Represents a pre-processed variable context that can be reused across multiple
+    expression evaluations for improved performance.
+
+    This class encapsulates the JSON round-trip safety processing, allowing you to
+    process variables once and reuse the safe representation multiple times.
+    """
+
+    def __init__(self, variables: dict | str | None = None):
+        """
+        Process variables into a safe, JSON-compatible format.
+
+        Args:
+            variables: Dictionary or JSON string containing variable values
+        """
+        self.processed_json = _process_variables(variables)
+
+    def get_processed_json(self) -> dict:
+        """Return the processed JSON object."""
+        return self.processed_json
+
+
 # Helper function to process variables
 def _process_variables(variables: dict | str | None = None) -> dict:
     """
@@ -436,6 +509,24 @@ def _process_variables(variables: dict | str | None = None) -> dict:
     return processed_json
 
 
+def _extract_processed_json(
+    context: Union[dict, str, "ProcessedContext", None] = None,
+) -> dict:
+    """
+    Extract processed JSON from various context types.
+
+    Args:
+        context: Raw variables, JSON string, or ProcessedContext instance
+
+    Returns:
+        Processed JSON object ready for expression evaluation
+    """
+    if isinstance(context, ProcessedContext):
+        return context.get_processed_json()
+    else:
+        return _process_variables(context)
+
+
 def compile_expression(expression: str) -> CompiledExpression:
     """
     Compile an expression into a reusable CompiledExpression object that can be
@@ -458,17 +549,20 @@ def compile_expression(expression: str) -> CompiledExpression:
 
 # Function to evaluate expressions
 def evaluate(
-    expression: str, variables: dict | str | None = None
-) -> int | float | bool | str:
+    expression: str, context: Union[dict, str, "ProcessedContext", None] = None
+) -> Union[int, float, bool, str]:
     """
-    Evaluate an expression with integers, arithmetic operations, comparisons,
-    and variables.
+    Evaluate an expression against the given context.
+
+    Args:
+        expression: The expression string to evaluate
+        context: Dictionary, JSON string, or ProcessedContext containing variable values
 
     For better performance when evaluating the same expression multiple times with
-    different variable contexts, use the compile() function instead.
+    different variable contexts, use compile_expression().
     """
 
-    processed_json = _process_variables(variables)
+    processed_json = _extract_processed_json(context)
     parser = build_parser()
 
     with parsing_error_handling(expression, parser.parse):
